import gc
import mock
import six
import types
import weakref

omitted = object()


def make_patches(target, make_wrapper, patch_all_referrers=True):
    """Return a list of mock._patch objects which wrap the given target function.

    The `target` argument must refer to the function you wish to patch:
    either a string: the dotted import path to the function, or a 2-tuple:
    an object and the name of one of its attributes.

    The `make_wrapper` argument must be a function that takes an initial
    `base` argument and returns that function, wrapped with whatever
    functionality you like.

    If `patch_all_referrers` is True (the default), then not only the given
    reference will be patched, but all other references to the same function
    object will be patched, if possible. This includes module and class
    members, as well as entries in dictionaries (like function registries).
    One important reference that CANNOT be patched is function closures;
    that is, if you have already wrapped function A with B in the typical
    functional way:

        def B(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                do_something()
                return func(*args, **kwargs)
            return wrapper

    ...then that internal call to func() will not be patched. It's not mutable.
    You can fix this by changing the wrapper function to a wrapper _class_
    which sets `self.func = func` and whose `__call__` method calls `self.func()`,
    because that can be discovered and patched.
    """
    if isinstance(target, six.string_types):
        primary_patch = mock.patch(target)
        original, local = primary_patch.get_original()
    elif isinstance(target, tuple) and len(target) == 2:
        if not isinstance(target[1], six.text_type):
            raise TypeError(
                "Targets which are (obj, funcname) 2-tuples MUST pass the funcname as a string."
            )
        primary_patch = mock.patch.object(*target)
        original, local = primary_patch.get_original()
    elif isinstance(target, (types.FunctionType, types.MethodType)):
        primary_patch = _patch_one(target)
        if primary_patch is None:
            raise TypeError("Cannot patch: %s could not be found." % (repr(target),))
        original = target
    else:
        raise TypeError(
            "Cannot patch: %s is not an (obj, attr) pair nor a dotted path name."
            % (repr(target),)
        )

    # Replace the target with a wrapper.
    if isinstance(
        original,
        (
            types.FunctionType,
            types.MethodType,
            types.BuiltinFunctionType,
            types.BuiltinMethodType,
        ),
    ):
        base = original
    elif isinstance(original, (staticmethod, classmethod)):
        base = original.__func__
    elif isinstance(original, property):
        base = original.fget
    else:
        raise TypeError("Cannot patch: %s is not a function." % (repr(target),))

    wrapper = make_wrapper(base)

    if isinstance(original, property):
        # We can't patch original.fget directly because it's read-only,
        # so we replace the whole property with a new one, passing our
        # wrapper as its fget.
        # At this time, we only patch fget. If there's enough demand,
        # we could do all three in the future, but then that would take
        # three wrapper functions, and what the instruments do
        # with three instead of one could be very confusing.
        primary_patch.new = property(
            wrapper, original.fset, original.fdel, original.__doc__
        )
    else:
        if isinstance(original, staticmethod):
            wrapper = staticmethod(wrapper)
        elif isinstance(original, classmethod):
            wrapper = classmethod(wrapper)
        primary_patch.new = wrapper

    patches = [primary_patch]

    if patch_all_referrers:
        # Add patches for any other modules/classes which have
        # the target as an attribute, or "registry" dicts which have
        # the target as a value.
        _resolved_target = primary_patch.getter()
        refs = gc.get_referrers(original)
        for ref in refs:
            if not isinstance(ref, dict):
                continue

            names = [k for k, v in ref.items() if v is original]
            seen_names = set()
            for parent in gc.get_referrers(ref):
                if parent is _resolved_target or parent is primary_patch:
                    continue
                if parent is wrapper:
                    # In Python 3.2+, `@functools.wraps(base)` above sets
                    # `wrapper.__wrapped__ = wrapped`. We don't want to
                    # patch that with itself!
                    continue

                if getattr(parent, "__dict__", None) is ref:
                    # An attribute of a "parent" module or class or instance.
                    for name in names:
                        patches.append(WeakMethodPatch(parent, name, wrapper))
                else:
                    for gpa in gc.get_referrers(parent):
                        if getattr(gpa, "__dict__", None) is parent:
                            # A member of a "parent" dict which is an attribute
                            # of a "grandparent" module or class or instance.
                            # ref[name] = original, where gpa.parent = ref
                            for name in names:
                                if name in seen_names:
                                    # Don't patch the same dict twice, or
                                    # a) we'll waste cycles, and
                                    # b) DictPatch.stop() may restore a patch
                                    # instead of the correct original.
                                    pass
                                else:
                                    patches.append(DictPatch(ref, name, wrapper))
                                    seen_names.add(name)
                            break

    return patches


def _patch_one(original):
    original_qualname = original.__qualname__

    # Try to find the original object
    refs = gc.get_referrers(original)
    for ref in refs:
        if not isinstance(ref, dict):
            continue

        names = [k for k, v in ref.items() if v is original]
        for parent in gc.get_referrers(ref):
            pq = (
                parent.__qualname__
                if isinstance(parent, type)
                else parent.__class__.__qualname__
            )
            # An attribute of a "parent" module or class or instance.
            for name in names:
                if "%s.%s" % (pq, name) == original_qualname:
                    return mock.patch.object(parent, name)


# ----------------------------- Weak patch ----------------------------- #


class WeakMethodPatch(object):
    """A Patch for an attribute a Python object.

    On start/__enter__, calls self.getter() which should return an object,
    then replaces the given attribute of that object with the new value.
    On stop/__exit__, replaces the same attribute with the previous value.

    Used by make_patches to replace references to functions which appear in
    modules, classes, or other objects. Weak references are used internally
    so that, if the object is removed from that module etc (has no more strong
    references), then the patch is automatically abandoned.
    """

    def __init__(self, obj, attribute, new):
        try:
            getter = weakref.ref(obj, self._safe_stop)
        except TypeError:

            def getter():
                return obj

        self.getter = getter
        self.attribute = attribute
        self.new = new

    def __repr__(self):
        return "%s(%s, %s, %s)" % (
            self.__class__.__name__,
            self.getter,
            self.attribute,
            self.new,
        )

    def _safe_stop(self, ref=None):
        try:
            self.stop()
        except RuntimeError:
            # Already stopped. Ignore.
            pass

    def get_original(self):
        target = self.getter()
        name = self.attribute

        original = omitted
        local = False

        try:
            original = target.__dict__[name]
        except (AttributeError, KeyError):
            original = getattr(target, name, omitted)
        else:
            local = True

        if original is omitted:
            raise AttributeError("%s does not have the attribute %r" % (target, name))
        return original, local

    def __enter__(self):
        """Perform the patch."""
        obj = self.getter()
        if obj is None:
            # The object we wanted to patch has already been garbage-collected.
            return

        original, local = self.get_original()
        self.temp_original = weakref.ref(original)
        self.is_local = local
        setattr(obj, self.attribute, self.new)
        return self.new

    def __exit__(self, *exc_info):
        """Undo the patch."""
        if not hasattr(self, "is_local"):
            raise RuntimeError("stop called on unstarted patcher")

        target = self.getter()
        if target is None:
            return

        original = self.temp_original()
        if original is None:
            return

        if getattr(target, self.attribute, None) is self.new:
            if self.is_local:
                setattr(target, self.attribute, original)
            else:
                delattr(target, self.attribute)
                if not hasattr(target, self.attribute):
                    # needed for proxy objects like django settings
                    setattr(target, self.attribute, original)

        del self.is_local

    def start(self):
        """Activate a patch, returning any created mock."""
        result = self.__enter__()
        return result

    def stop(self):
        """Stop an active patch."""
        return self.__exit__()


class DictPatch(object):
    """A Patch for a member of a Python dictionary.

    On start/__enter__, replaces the member of the given dictionary
    identified by the given key with a new object. On stop/__exit__,
    replaces the same key with the previous object.

    Used by make_patches to replace references to functions which appear
    in any dictionary, such as a function registry.
    """

    def __init__(self, dictionary, key, new):
        self.dictionary = dictionary
        self.key = key
        self.new = new

    def get_original(self):
        return self.dictionary[self.key], True

    def __enter__(self):
        """Perform the patch."""
        original, local = self.get_original()
        self.temp_original = original
        self.is_local = local
        self.dictionary[self.key] = self.new
        return self.new

    def __exit__(self, *exc_info):
        """Undo the patch."""
        if not hasattr(self, "is_local"):
            raise RuntimeError("stop called on unstarted patcher")

        self.dictionary[self.key] = self.temp_original

        del self.is_local

    def start(self):
        """Activate a patch, returning any created mock."""
        result = self.__enter__()
        return result

    def stop(self):
        """Stop an active patch."""
        return self.__exit__()



"""Wrappers to monitor function execution."""

from collections import namedtuple
import datetime
import functools
import linecache
import sys
import time
import traceback

import hunter
import six

from diagnose import patchlib

omitted = object()

# A registry of FunctionProbe instances.
# Since probes patch functions, it's imperative that the same function
# not be patched twice, even in test suites or other scenarios where
# two services might be applying probes at the same time. This global
# registry enforces that. Callers should add instruments to probes instead.
# The FunctionProbe class itself uses this to disallow attaching two
# FunctionProbes to the same target.
active_probes = {}


def attach_to(target):
    """Return the probe attached to the given target, or a new one if needed."""
    probe = active_probes.get(target, None)
    if probe is None:
        probe = FunctionProbe(target)
        active_probes[target] = probe
    return probe


def start_all():
    for probe in six.itervalues(active_probes):
        probe.start()


class FunctionProbe(object):
    """A wrapper for a function, to monitor its execution.

    target: a dotted Python path to the function to wrap.
    instruments: a dict of ("spec id": Instrument instance) pairs.

    When started, a FunctionProbe "monkey-patches" its target, replacing
    it with a wrapper function. That wrapper calls the original target
    function, calling each instrument at the event it specifies: call, end,
    or return. If the event is "end", then the probe registers a trace function
    (via `sys.settrace`) and then calls those instruments just before the
    original target function returns; the instrument values are evaluated
    in the globals and locals of that frame. "Call" and "return" instruments
    are evaluated before or after the function returns; they are much less
    invasive since they do not call settrace, but they are limited to reading
    the function's inputs, outputs, and timing.
    """

    def __init__(self, target, instruments=None):
        if target in active_probes:
            raise RuntimeError(
                "Cannot apply two probes to the same target. "
                "Try calling attach_to(target) instead of FunctionProbe(target)."
            )
        self.target = target
        if instruments is None:
            instruments = {}
        self.instruments = instruments
        self.patches = []
        active_probes[target] = self

    def __str__(self):
        return "%s(target=%r, instruments=%r)" % (
            self.__class__.__name__,
            self.target,
            self.instruments,
        )

    __repr__ = __str__

    def make_wrapper(self, base):
        varnames = self.maybe_unwrap(base).__code__.co_varnames

        @functools.wraps(base)
        def probe_wrapper(*args, **kwargs):
            now = datetime.datetime.utcnow()

            hotspots = HotspotsFinder()
            instruments_by_event = {"call": [], "return": [], "end": []}
            for I in six.itervalues(self.instruments):
                if I.expires and now > I.expires:
                    continue
                if I.check_call(self, *args, **kwargs):
                    instruments_by_event[I.event].append(I)
                    if I.event in ("call", "return"):
                        if "hotspots" in I.value or "hotspots" in (
                            I.custom.get("tags") or ""
                        ):
                            hotspots.enabled = True

            target_obj, target_func_name = self.target.rsplit(".", 1)
            is_unwrapped = base.__code__.co_name == target_func_name
            if instruments_by_event["end"]:
                # We have instruments that require evaluation in the local
                # context of the function. Call sys.settrace() to gain access.
                predicate = hunter.When(
                    hunter.Query(
                        # Only trace returns (this will include exceptions!)...
                        kind="return",
                        # ...and only in the given function...
                        function=target_func_name,
                        # ...(no deeper).
                        depth=0,
                    )
                    if is_unwrapped
                    else hunter.Query(
                        # Only trace returns (this will include exceptions!)...
                        kind="return",
                        # ...and only in the given function...
                        function=target_func_name,
                        # ...but we don't know how many times it's been wrapped.
                        # Use the module instead as an approximate match.
                        # This may catch other functions with the same name
                        # in the same module, but not much we can do about
                        # that without a custom Cython Query.
                        module_in=target_obj,
                    ),
                    TraceHandler(self, instruments_by_event["end"]),
                )
                tracer = hunter.Tracer(
                    # There's no need to call threading.settrace() because
                    # a) we're targeting a function we're about to call
                    #    in the same thread,
                    # b) we're going to undo it immediately after, and
                    # c) it would collide with other threads if they did
                    #    the same concurrently.
                    threading_support=False
                ).trace(predicate)
            elif hotspots.enabled:
                # We have instruments that require timing internal lines.
                # Call sys.settrace() to gain access.
                predicate = hunter.When(
                    hunter.Query(
                        # Only trace lines...
                        kind="line",
                        # ...and only in the given function...
                        function=target_func_name,
                        # ...(no deeper).
                        depth=1,
                    )
                    if is_unwrapped
                    else hunter.Query(
                        # Only trace lines...
                        kind="line",
                        # ...and only in the given function...
                        function=target_func_name,
                        # ...but we don't know how many times it's been wrapped.
                        # Use the module instead as an approximate match.
                        # This may catch other functions with the same name
                        # in the same module, but not much we can do about
                        # that without a custom Cython Query.
                        module_in=target_obj,
                    ),
                    hotspots,
                )
                tracer = hunter.Tracer(
                    # There's no need to call threading.settrace() because
                    # a) we're targeting a function we're about to call
                    #    in the same thread,
                    # b) we're going to undo it immediately after, and
                    # c) it would collide with other threads if they did
                    #    the same concurrently.
                    threading_support=False
                ).trace(predicate)
            else:
                tracer = None

            try:
                if instruments_by_event["call"] or instruments_by_event["return"]:
                    start = time.time()
                    _locals = {
                        "start": start,
                        "now": now,
                        "args": args,
                        "kwargs": kwargs,
                        "frame": sys._getframe(),
                    }
                    # Add positional args to locals by name.
                    for i, argname in enumerate(varnames[: len(args)]):
                        if argname not in ("args", "kwargs"):
                            _locals[argname] = args[i]
                    # Add kwargs to locals
                    _locals.update(kwargs)

                for instrument in instruments_by_event["call"]:
                    try:
                        instrument.fire(instrument.mgr.global_namespace, _locals)
                    except:
                        try:
                            instrument.handle_error(self)
                        except:
                            traceback.print_exc()

                # Execute the base function and obtain its result.
                try:
                    result = base(*args, **kwargs)
                except:
                    result = sys.exc_info()[1]
                    raise
                finally:
                    if hotspots.enabled:
                        hotspots.finish()
                        _locals["hotspots"] = hotspots

                    if instruments_by_event["return"]:
                        end = time.time()
                        elapsed = end - start
                        _locals.update(
                            {"result": result, "end": end, "elapsed": elapsed}
                        )

                    for instrument in instruments_by_event["return"]:
                        try:
                            instrument.fire(instrument.mgr.global_namespace, _locals)
                        except:
                            try:
                                instrument.handle_error(self)
                            except:
                                traceback.print_exc()

                return result
            finally:
                if tracer is not None:
                    tracer.stop()

                for I in six.itervalues(self.instruments):
                    I.finish()

        return probe_wrapper

    @staticmethod
    def maybe_unwrap(func):
        """Return the given function, without its probe_wrapper if it has one."""
        if getattr(getattr(func, "__code__", None), "co_name", "") == "probe_wrapper":
            return func.__closure__[0].cell_contents
        else:
            try:
                # If the given func is a func returned from @functools.wraps(orig),
                # then the only cell in its closure will be the `orig` function.
                # There may be funcs with a single cell that is a function
                # that is not the wrapped function, in which case we will
                # return the wrong signature here, but use of wraps is so
                # much more common and SO useful to unwrap that we risk it.
                if len(func.__closure__) == 1:
                    f = func.__closure__[0].cell_contents
                    if hasattr(f, "__code__"):
                        return f
            except (AttributeError, IndexError, TypeError):
                pass

        return func

    def start(self):
        """Apply self.patches. Safe to call after already started."""
        if not self.patches:
            self.patches = patchlib.make_patches(self.target, self.make_wrapper)
        for p in self.patches:
            if not hasattr(p, "is_local"):
                p.start()

    def stop(self):
        for p in self.patches:
            try:
                p.stop()
            except RuntimeError:
                # Already stopped. Ignore.
                pass


class TraceHandler(object):
    """A sys.settrace arg, which calls instruments in the context of the frame."""

    def __init__(self, probe, instruments):
        self.probe = probe
        self.instruments = instruments

    def __call__(self, event):
        _globals = event.globals
        _locals = {"__event__": event}
        _locals.update(event.locals)
        for instrument in self.instruments:
            try:
                _g = _globals.copy()
                _g.update(instrument.mgr.global_namespace)
                instrument.fire(_g, _locals)
            except:
                try:
                    instrument.handle_error(self.probe)
                except:
                    traceback.print_exc()


CallTime = namedtuple("CallTime", ["time", "lineno", "source"])


class HotspotsFinder(object):
    """A sys.settrace arg, which records line timings."""

    def __init__(self):
        self.enabled = False
        self._last_time = None
        self._last_line = None
        self.calls = {}
        self.filename = None

    def __call__(self, event=None):
        if self._last_time is not None:
            elapsed = time.time() - self._last_time
            ll = self._last_line
            call = self.calls.get(ll, None)
            if call is None:
                # count, max, sum
                self.calls[ll] = [1, elapsed, elapsed]
            else:
                call[0] += 1
                if elapsed > call[1]:
                    call[1] = elapsed
                call[2] += elapsed

        if event is not None:
            self._last_line = event.lineno
            if self.filename is None:
                self.filename = event.filename
            # Don't include this method's time in the next line time
            self._last_time = time.time()

    def finish(self):
        # Fake the last line time
        self.__call__(event=None)

        if self.calls:
            worst = max(
                ((call[2], lineno) for lineno, call in six.iteritems(self.calls))
            )
            self.worst = CallTime(*worst, source=self.source(worst[1]))

            slowest = max(
                ((call[1], lineno) for lineno, call in six.iteritems(self.calls))
            )
            self.slowest = CallTime(*slowest, source=self.source(slowest[1]))
        else:
            self.worst = self.slowest = CallTime(None, None, None)

    def source(self, lineno):
        return linecache.getline(self.filename, lineno)



"""A tool to detect and synchronize execution or simulate errors during tests.

Use a Breakpoint to detect concurrent calls, simulate deterministic
interleavings of operations, or throw internal errors which would
otherwise be left to concurrency accidents or complicated calls
to sleep(). All of this can be done entirely in tests, without infecting
production code with a bunch of scaffolding.

When a Breakpoint is hit, its condition is checked (if any), and the True
or False result is appended to its .calls. Callpoints without a condition
append True every time they are called. Test code may call bp.wait(),
which blocks until there has been at least one (conditional) call. For example:

    with Breakpoint("path.to.obj.func", event="return") as bp:
        bp.start_thread(foo)  # start running foo in a thread
        bp.wait()             # wait for that thread to hit the function

        assert some_effect_that_func_causes()

If the system somehow hits the breakpoint before the test calls wait(),
then wait() simply returns immediately.
If no calls happen within the timeout, wait() throws a RuntimeError.
If test code does not call wait(), and the "with" block exits with no calls
recorded, it will throw a RuntimeError. You can use this to assert calls even
in completely synchronous code.

Breakpoints may be conditional based on the arguments being passed.
Set bp.condition to a callable which takes the same arguments as the
target function; if it returns True, the call is recorded as successful.
Alternately, the condition may be an int or list of ints, in which case
it will be considered successful on those numbered calls (starting from 0).

When a breakpoint is hit, the breakpoint's "stackframe" attribute is
set to the current frame. Using this, you can inspect the call
stack or function arguments while inside the "with" block.

Blocking Breakpoints
--------------------

A blocking Breakpoint does not merely record calls, but blocks the caller
on successful calls until released. This lets it behave like a semaphore,
allowing a test and the system (the thread being tested) to synchronize
their execution. The test calls bp.wait(), which blocks; when the system
hits the breakpoint, they switch: the system blocks and the test unblocks.
While the system is blocked, the test is free to inspect or alter state
without fear.

For example:

    with Breakpoint.block((obj, funcname)) as bp:
        bp.start_thread(foo)  # start running foo in a thread
        bp.wait()             # wait for that thread to hit the function

        assert foo.is_running
        foo.do_something()

The test may unblock the system by explicitly calling bp.release(),
or just waiting for the "with" block to exit. If neither happens
within the timeout, the system thread throws RuntimeError.

Erroring Breakpoints
--------------------

An erroring Breakpoint is initialized with an exception instance, which is raised
whenever the condition is met. Use this to simulate errors inside concurrent
code at specific points.

For example:

    with Breakpoint.error("path.to.obj.func") as bp:
        bp.start_thread(foo)  # start running foo in a thread
        bp.wait()             # wait for that thread to hit the function

        assert "server thread failed" in error_logs

The `do` class
--------------

Breakpoints have more power, but can be confusing because you have to
declare what to run after you declare where it should block. The `do` class
allows you to reverse that order, and use a more English-like chain of
adjustments, with less boilerplate. It assumes you want to do the most
common thing, which is start a thread running some function and then wait
for it to hit the Breakpoint. For example, instead of:

    with Breakpoint("path.to.obj.func", event="return") as bp:
        bp.start_thread(foo)
        bp.wait()
        etc

...this class lets you write:

    with do(foo).until("path.to.obj.func").returns:
        etc

Entering the context starts a thread running the given function and will
wait() for the Breakpoint to be hit. Exiting the context will join()
the thread.

Most do() methods and properties return the do() instance itself, so it doesn't
matter which one comes last; they can all be passed to a `with` block, like so:

    t = do(foo)
    with t.until("path.to.obj.func").returns:
        etc

This is especially handy for another common case: gathering results. Every do()
instance stores the output in self.results automatically, even if you call it
more than once:

    inputs = ["a", "b"]
    t = do(len, inputs)
    with t:
        pass
    inputs.append("c")
    with t:
        pass
    assert t.results == [2, 3]

"""

import functools
import inspect
import threading
import time

import six

from diagnose import patchlib


omitted = object()


class Breakpoint:
    """A tool to detect and synchronize execution or simulate errors during tests."""

    check_interval = 0.1
    """The period, in seconds, at which to poll for timeout."""

    patch_all_referrers = None
    """If True, all references to the given target will be patched.
    If False, only the given reference will be patched. If None
    (the default), all references will be patched when `target`
    is a string (dotted-import path), but only the given reference
    will be patched if `target` is an (object, attribute-name) tuple."""

    def __init__(
        self, target, event="call", condition=None, timeout=10.0, fire=None,
    ):
        """
        target:
            Function to be patched: may be the dotted-import path as a string
            or an (obj, funcname) tuple.
        event:
            "call" (default) to fire when the function is entered,
            "return" to fire when the function exits,
            "error" to fire when the function throws an Exception.
        condition:
            None to always fire, or a callable that takes the same args
            as the patched target and returns True to fire, False to not.
            Alternately, it may be an int or list of ints, in which case
            it will fire on those numbered calls.
        timeout:
            The default time, in seconds, to wait().
        fire:
            None (the default) to take no action when the Breakpoint is hit.
            Pass a no-arg callable to perform some other action.
            Use Breakpoint.block() or Breakpoint.error() to use
            the builtin actions.
        """
        self.target = target
        self.event = event
        self.condition = condition
        self.timeout = timeout

        self._started_threads = []
        self.stackframe = None

        self.calls = []
        self.hits = 0
        self.fire = fire

    def _make_wrapper(self, base):
        """A function wrpper which fires any internal action for a Breakpoint."""

        @functools.wraps(base)
        def breakpoint_wrapper(*args, **kwargs):
            self.stackframe = inspect.currentframe()
            if self.event == "call":
                if self._condition_met(args, kwargs):
                    if self.fire is not None:
                        self.fire()

            try:
                result = base(*args, **kwargs)
            except Exception:
                if self.event == "error":
                    if self._condition_met(args, kwargs):
                        if self.fire is not None:
                            self.fire()
                raise
            else:
                if self.event == "return":
                    if self._condition_met(args, kwargs):
                        if self.fire is not None:
                            self.fire()
            finally:
                # Best practice is not to hold onto stackframe longer
                # than it is needed.
                self.stackframe = None

            return result

        return breakpoint_wrapper

    def _condition_met(self, args, kwargs):
        if self.condition is None:
            met = True
        elif isinstance(self.condition, int):
            met = len(self.calls) == self.condition
        elif isinstance(self.condition, (set, tuple, list)):
            met = len(self.calls) in self.condition
        elif callable(self.condition):
            met = self.condition(*args, **kwargs)
        else:
            raise TypeError(
                "Breakpoint.condition must be None, an int or list of ints, or a callable."
            )
        self.calls.append(met)
        if met:
            self.hits += 1
        return met

    def __enter__(self):
        self.calls = []
        self.hits = 0
        self._started_threads = []
        self.release()

        if self.target is None:
            self.patches = []
        else:
            patch_all = self.patch_all_referrers
            if patch_all is None:
                patch_all = isinstance(self.target, six.string_types)
            self.patches = patchlib.make_patches(
                self.target, self._make_wrapper, patch_all_referrers=patch_all
            )

        for p in self.patches:
            p.start()

        return self

    def __exit__(self, type, value, traceback):
        while self.patches:
            p = self.patches.pop(0)
            p.stop()

        self.release()

        while self._started_threads:
            t = self._started_threads.pop(0)
            t.join()

        if type is not None:
            # There was already an error, don't suppress it.
            return False

        if self.target is not None:
            if isinstance(self.condition, int):
                if len(self.calls) <= self.condition or not self.calls[self.condition]:
                    raise AssertionError(
                        "Breakpoint condition on %s was not met for iteration %s."
                        % (self.target, self.condition)
                    )
            elif isinstance(self.condition, (set, tuple, list)):
                not_called = [
                    c
                    for c in self.condition
                    if len(self.calls) <= c or not self.calls[c]
                ]
                if not_called:
                    raise AssertionError(
                        "Breakpoint condition on %s was not met for iterations %s."
                        % (self.target, not_called)
                    )
            else:
                if not any(self.calls):
                    raise AssertionError(
                        "Breakpoint condition on %s was not met." % (self.target,)
                    )

    def start_thread(self, func, *args, **kwargs):
        """Execute the given func (a callable) in another thread.

        Tests may call this to run the system in the background.

        Threads started here will be joined on breakpoint context exit.
        """
        t = threading.Thread(
            target=func,
            name="Breakpoint_%s_%s" % (func.__name__, len(self._started_threads)),
            args=args,
            kwargs=kwargs,
        )
        self._started_threads.append(t)
        t.start()

    def join(self):
        """Wait for all threads started by this instance to complete."""
        for thread in self._started_threads:
            thread.join()

    def wait(self, timeout=omitted, hits=1):
        """Block until the breakpoint is hit by other thread(s).

        Tests should call this to wait for other threads to call the
        breakpoint.target function. Once they do, this method returns and the
        test may proceed.

        If `hits` is greater than 1, this function waits until the breakpoint
        has been hit the given number of times before returning.
        """
        if self.target is None:
            return

        if timeout is omitted:
            timeout = self.timeout

        start = time.time()
        while self.hits < hits:
            if timeout is not None and time.time() - start > timeout:
                raise RuntimeError(
                    "Breakpoint on %s (event='%s') not hit after %s seconds."
                    % (self.target, self.event, timeout)
                )
            time.sleep(self.check_interval)

    def wait_until(self, condition, timeout=omitted):
        """Block until the condition is True, or error if the timeout is reached.

        Call this to wait for the system to advance to a certain state;
        once the given `condition` callable returns True, this function
        returns. If the timeout (or self.timeout if omitted) is reached,
        a RuntimeError is raised.
        """
        if timeout is omitted:
            timeout = self.timeout

        start = time.time()
        while not condition():
            if timeout is not None and time.time() - start > timeout:
                raise RuntimeError(
                    "Condition for %s not met after %s seconds."
                    % (self.target, timeout)
                )
            time.sleep(self.check_interval)

    # ------------------------- Blocking breakpoints ------------------------- #

    @classmethod
    def block(cls, *args, **kwargs):
        """Create a Breakpoint which blocks, to synchronize execution."""
        self = cls(*args, **kwargs)
        self.fire = self._fire_blocking
        return self

    def _fire_blocking(self):
        """The internal action for a blocking Breakpoint."""
        start = time.time()
        timeout = self.timeout
        self.blocked = True
        while self.blocked:
            if timeout is not None and time.time() - start > timeout:
                raise RuntimeError(
                    "Breakpoint on %s timed out after %s seconds."
                    % (self.target, timeout)
                )
            time.sleep(self.check_interval)

    def release(self):
        """Allow the system to proceed (until the breakpoint is hit again).

        When the system hits a blocking breakpoint, it blocks while the test proceeds.
        When the context exits, it calls this method to unblock the system.
        If the test wants to unblock the system before the context exits,
        it may call this method directly.
        """
        self.hits = 0
        self.blocked = False

    # ------------------------- Erroring breakpoints ------------------------- #

    @classmethod
    def error(cls, exception, *args, **kwargs):
        """Create a Breakpoint which throws the given exception.

        The `exception` argument must be an exception instance, which is raised
        whenever the condition is met.
        """
        self = cls(*args, **kwargs)
        self.exception = exception
        self.fire = self._fire_erroring
        return self

    def _fire_erroring(self):
        """The internal action for an erroring Breakpoint."""
        raise self.exception


class do:
    """A helper class to run concurrent functions with Breakpoints."""

    breakpoint = None

    def __init__(self, func=None, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.breakpoint = Breakpoint(None)
        self.results = []

    def until(self, target, timeout=None):
        """Set a blocking Breakpoint for the given target."""
        self.breakpoint.target = target
        self.breakpoint.fire = self.breakpoint._fire_blocking
        if timeout is not None:
            self.breakpoint.timeout = timeout
        return self

    def beyond(self, target, timeout=None):
        """Set a non-blocking Breakpoint for the given target."""
        self.breakpoint.target = target
        if timeout is not None:
            self.breakpoint.timeout = timeout
        return self

    def error_on(self, target, exception, timeout=None):
        """Set an erroring Breakpoint for the given target."""
        self.breakpoint.target = target
        self.breakpoint.exception = exception
        self.breakpoint.fire = self.breakpoint._fire_erroring
        if timeout is not None:
            self.breakpoint.timeout = timeout
        return self

    @property
    def returns(self):
        """Set the Breakpoint to fire when the target returns, not when called."""
        if self.breakpoint is None:
            raise RuntimeError(
                "You must call do().until(), .beyond(), or .error_on() before declaring .returns."
            )
        self.breakpoint.event = "return"
        return self

    @property
    def errors(self):
        """Set the Breakpoint to fire when the target errors, not when called."""
        if self.breakpoint is None:
            raise RuntimeError(
                "You must call do().until(), .beyond(), or .error_on() before declaring .errors."
            )
        self.breakpoint.event = "error"
        return self

    def where(self, condition):
        """Set the Breakpoint to fire only when the given condition is met."""
        self.breakpoint.condition = condition
        return self

    @property
    def once(self):
        """Set the Breakpoint to fire only once."""
        self.breakpoint.condition = 0
        return self

    def __enter__(self):
        self.breakpoint.__enter__()
        if self.func is not None:
            self.breakpoint.start_thread(self._gather_results)
            self.breakpoint.wait()
        return self

    def __exit__(self, type, value, traceback):
        self.breakpoint.__exit__(type, value, traceback)

    def _gather_results(self):
        try:
            result = self.func(*self.args, **self.kwargs)
        except Exception as exc:
            result = exc
            raise
        finally:
            self.results.append(result)

    def release(self):
        """Release any threads blocked on the Breakpoint."""
        self.breakpoint.release()



from setuptools import find_packages, setup

ext_modules = []

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="diagnose",
    version="4.0.0a3",
    author="Robert Brewer",
    author_email="dev@crunch.io",
    description="A library for instrumenting Python code at runtime.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Crunch-io/diagnose",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=["hunter>=2.2.0", "mock", "six"],
    ext_modules=ext_modules,
    entry_points={},
)



import functools
from six.moves import xrange

import diagnose


def wrap(f):
    class Other(object):
        def __init__(self, f):
            self.f = f

    other = Other(f)

    @functools.wraps(f)
    def inner(self, arg1):
        return other.f(self, arg1 + 100)

    return inner


def a_func(arg):
    extra = 13
    output = arg + extra
    return output


def func_2(arg):
    extra = 17
    output = arg + extra
    return output


def hard_work(lower, upper):
    output = xrange(lower, upper)
    summary = len([x for x in output if x % 10 == 0])
    return summary


def to_columns(rowlist):
    """Convert [{A: 1, B: 5}, {A: 2, B: 6}] to {A: [1, 2], B: [5, 6]}."""
    dictcols = {}
    for i, rowobj in enumerate(rowlist):
        for k, v in rowobj.items():
            dictcols.setdefault(k, [None] * i).append(v)
    for k, v in dictcols.items():
        if len(v) < i:
            dictcols[k] += [None] * (i - len(v))
    return dictcols


class Thing:

    stage = None

    def __init__(self, template="<%s>"):
        self.template = template

    def do(self, arg, user_id=None):
        return self.template % arg

    @staticmethod
    def static():
        return 15

    @wrap
    def add5(self, arg1):
        return arg1 + 5

    @property
    def exists(self):
        return True

    def advance_stage(self, to):
        self.stage = to

    def advance_multiple(self):
        self.advance_stage(to="alpha")
        self.advance_stage(to="beta")
        self.advance_stage(to="gamma")
        self.advance_stage(to="delta")


class ClassDecorator(object):
    def __call__(self, fn):
        @functools.wraps(fn)
        def fn2(*args, **kwargs):
            return fn(*args, **kwargs)

        return fn2


@ClassDecorator()
def sum4(arg1, arg2, arg3, arg4):
    return arg1 + arg2 + arg3 + arg4


def orig(term):
    return term[:2] + "a!"


funcs = {"orig": orig}


@diagnose.instruments.ProbeTestInstrument("mult_by_8", "result")
def mult_by_8(arg):
    return arg * 8
